package de.arminwolf.configs;

import de.arminwolf.exceptions.AppException;
import de.arminwolf.models.User;
import de.arminwolf.models.dto.RoleDTO;
import de.arminwolf.models.dto.UserDTO;
import de.arminwolf.services.UserService;
import lombok.RequiredArgsConstructor;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.http.HttpStatus;
import org.springframework.security.authentication.AuthenticationManager;
import org.springframework.security.authentication.UsernamePasswordAuthenticationToken;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.AuthenticationException;
import org.springframework.security.oauth2.jwt.Jwt;
import org.springframework.security.oauth2.jwt.JwtClaimsSet;
import org.springframework.security.oauth2.jwt.JwtDecoder;
import org.springframework.security.oauth2.jwt.JwtEncoder;
import org.springframework.security.oauth2.jwt.JwtEncoderParameters;
import org.springframework.stereotype.Component;

import java.net.MalformedURLException;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.time.ZoneOffset;
import java.util.Date;
import java.util.Objects;
import java.util.Optional;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

@RequiredArgsConstructor
@Component
public class AuthenticationProvider {

	private static Logger LOG = LoggerFactory.getLogger(AuthenticationProvider.class);

	public static final String ROLES    = "roles";
	public static final String USERNAME = "username";
	@Value("${security.jwt.token.issuer:https://armin-wolf.de}")
	private             String issuer;


	private final UserService userService;
	private final JwtDecoder  jwtDecoder;
	private final JwtEncoder  jwtEncoder;


	public Authentication validateToken(final String token) {
		Jwt decodedToken = jwtDecoder.decode(token);
		// check validity of token
		if (Objects.isNull(decodedToken) || !isValid(decodedToken)) {
			throw new AppException("Invalid token", HttpStatus.UNAUTHORIZED);
		}

		Optional<User> byUsername = userService.findByUsername(decodedToken.getClaim(USERNAME));
		if (byUsername.isEmpty()) {
			throw new AppException("User not found", HttpStatus.BAD_REQUEST);
		} else {
			User user = byUsername.get();
			return new UsernamePasswordAuthenticationToken(user, null, user.getAuthorities());
		}
	}


	private boolean isValid(final Jwt decodedToken) {
		if (!Objects.equals(decodedToken.getIssuer().toString(), issuer)) {
			LOG.warn("Invalid issuer: '{}'.", decodedToken.getIssuer());
			return false;
		}

		if (Objects.isNull(decodedToken.getExpiresAt()) || decodedToken.getExpiresAt().isAfter(LocalDateTime.now().toInstant(ZoneOffset.UTC))) {
			LOG.warn("Token expired: '{}'.", decodedToken.getExpiresAt());
			return false;
		}

		return true;
	}


	public String createToken(final UserDTO user) {
		Date now      = new Date();
		Date validity = new Date(now.getTime() + TimeUnit.HOURS.toMillis(1));

		JwtClaimsSet jwtClaimsSet = JwtClaimsSet.builder()
				.issuer(issuer)
				.claim(USERNAME, user.getUsername())
				.claim(ROLES, user.getRoles().stream().map(RoleDTO::getName).collect(Collectors.joining(",")))
				.expiresAt(validity.toInstant())
				.issuedAt(now.toInstant())
				.build();

		LOG.info("Creating token for user: '{}'.", user.getUsername());
		JwtEncoderParameters jwtEncoderParameters = JwtEncoderParameters.from(jwtClaimsSet);
		return jwtEncoder.encode(jwtEncoderParameters).getTokenValue();
	}
}
